@@ -12,7 +12,7 @@
 
 import logging
 from enum import Enum
-from typing import Optional, Sequence
+from typing import Optional, Sequence, Dict, Any, Tuple
 
 import mlflow.pyfunc
 import mlflow.pytorch
@@ -22,6 +22,7 @@ from mlflow.protos.databricks_pb2 import ErrorCode, RESOURCE_DOES_NOT_EXIST
 from mlflow.tracking import MlflowClient
 
 from numalogic.registry import ArtifactManager, ArtifactData
+from numalogic.tools.exceptions import ModelVersionError
 from numalogic.tools.types import Artifact
 
 _LOGGER = logging.getLogger()
@@ -39,7 +40,10 @@ class ModelStage(str, Enum):
 
 class MLflowRegistry(ArtifactManager):
     """
-    Model saving and loading using MLFlow Registry.
+    Model saving and loading using MLFlow Registry. The parameter model_stage
+    determines what environment we are using. The old models are moved to
+    'Archived' state and the latest model comes to 'Staging' or 'Production'
+    depending on model_stage parameter.
 
     More details here: https://mlflow.org/docs/latest/model-registry.html
 
@@ -49,6 +53,9 @@ class MLflowRegistry(ArtifactManager):
                               supported values include:
                               {"pytorch", "sklearn", "tensorflow", "pyfunc"}
         models_to_retain: number of models to retain in the DB (default = 5)
+        model_stage: Staging environment from where to load the latest model from (mlflow )
+                            supported values include:
+                              {"Staging", "Production", "Archived"}(default = "Production")
 
     Examples
     --------
@@ -63,7 +70,7 @@ class MLflowRegistry(ArtifactManager):
     >>> artifact_data = registry.load(skeys=["model"], dkeys=["AE"])
     """
 
-    __slots__ = ("client", "handler", "models_to_retain")
+    __slots__ = ("client", "handler", "models_to_retain", "model_stage")
     _TRACKING_URI = None
 
     def __new__(
@@ -71,6 +78,7 @@ class MLflowRegistry(ArtifactManager):
         tracking_uri: Optional[str],
         artifact_type: str = "pytorch",
         models_to_retain: int = 5,
+        model_stage: ModelStage = ModelStage.PRODUCTION,
         *args,
         **kwargs,
     ):
@@ -80,13 +88,18 @@ class MLflowRegistry(ArtifactManager):
         return instance
 
     def __init__(
-        self, tracking_uri: str, artifact_type: str = "pytorch", models_to_retain: int = 5
+        self,
+        tracking_uri: str,
+        artifact_type: str = "pytorch",
+        models_to_retain: int = 5,
+        model_stage: str = ModelStage.PRODUCTION,
     ):
         super().__init__(tracking_uri)
         mlflow.set_tracking_uri(tracking_uri)
         self.client = MlflowClient()
         self.handler = self.mlflow_handler(artifact_type)
         self.models_to_retain = models_to_retain
+        self.model_stage = model_stage
 
     @staticmethod
     def construct_key(skeys: Sequence[str], dkeys: Sequence[str]) -> str:
@@ -127,25 +140,19 @@ class MLflowRegistry(ArtifactManager):
     ) -> Optional[ArtifactData]:
         model_key = self.construct_key(skeys, dkeys)
         try:
-            if latest:
-                model = self.handler.load_model(
-                    model_uri=f"models:/{model_key}/{ModelStage.PRODUCTION}"
-                )
-                version_info = self.client.get_latest_versions(
-                    model_key, stages=[ModelStage.PRODUCTION]
-                )[-1]
-            elif version is not None:
-                model = self.handler.load_model(model_uri=f"models:/{model_key}/{version}")
-                version_info = self.client.get_model_version(model_key, version)
+            if (latest and version) or (not latest and not version):
+                raise ValueError("Either One of 'latest' or 'version' needed in load method call")
+
+            elif latest:
+                version_info = self.client.get_latest_versions(model_key, stages=[self.model_stage])
+                if not version_info:
+                    raise ModelVersionError("Model version missing for key = %s" % model_key)
+                version_info = version_info[-1]
             else:
-                raise ValueError("One of 'latest' or 'version' needed in load method call")
-            _LOGGER.info("Successfully loaded model %s from Mlflow", model_key)
-
-            run_info = mlflow.get_run(version_info.run_id)
-            metadata = run_info.data.params or None
-            _LOGGER.info("Successfully loaded model metadata from Mlflow!")
-
+                version_info = self.client.get_model_version(model_key, version)
+            model, metadata = self.__load_artifacts(skeys, dkeys, version_info)
             return ArtifactData(artifact=model, metadata=metadata, extras=dict(version_info))
+
         except RestException as mlflow_err:
             if ErrorCode.Value(mlflow_err.error_code) == RESOURCE_DOES_NOT_EXIST:
                 _LOGGER.info("Model not found with key: %s", model_key)
@@ -154,8 +161,15 @@ class MLflowRegistry(ArtifactManager):
                     "Mlflow error when loading a model with key: %s: %r", model_key, mlflow_err
                 )
             return None
+        except ModelVersionError as model_missing_err:
+            _LOGGER.error(
+                "No Model found found in %s ERROR: %r",
+                self.model_stage,
+                model_missing_err,
+            )
+            return None
         except Exception as ex:
-            _LOGGER.exception("Error when loading a model with key: %s: %r", model_key, ex)
+            _LOGGER.exception("Unexpected error: %s", ex)
             return None
 
     def save(
@@ -226,48 +240,56 @@ class MLflowRegistry(ArtifactManager):
         """
         model_name = self.construct_key(skeys, dkeys)
         try:
-            version = int(self.get_version(model_name=model_name))
+            current_staging = self.client.get_latest_versions(
+                name=model_name, stages=[self.model_stage]
+            )
+            latest = self.client.get_latest_versions(name=model_name, stages=["None"])
+
             latest_model_data = self.client.transition_model_version_stage(
                 name=model_name,
-                version=str(version),
-                stage=ModelStage.PRODUCTION,
+                version=str(latest[-1].version),
+                stage=self.model_stage,
             )
-            if version - 1 > 0:
+
+            if current_staging:
                 self.client.transition_model_version_stage(
                     name=model_name,
-                    version=str(version - 1),
-                    stage=ModelStage.STAGE,
-                )
-            if version - 2 > 0:
-                self.client.transition_model_version_stage(
-                    name=model_name,
-                    version=str(version - 2),
+                    version=str(current_staging[-1].version),
                     stage=ModelStage.ARCHIVE,
                 )
 
             # only keep "models_to_retain" number of models.
-            list_model_versions = list(self.client.search_model_versions(f"name='{model_name}'"))
-            models_to_delete = list_model_versions[: -self.models_to_retain]
-            for stale_model in models_to_delete:
-                self.delete(skeys=skeys, dkeys=dkeys, version=stale_model.version)
+            self.__delete_stale_models(skeys=skeys, dkeys=dkeys)
+
             _LOGGER.info("Successfully transitioned model to Production stage")
             return latest_model_data
-        except Exception as ex:
+        except RestException as ex:
             _LOGGER.exception(
                 "Error when transitioning a model: %s to different stage: %r", model_name, ex
             )
             return None
 
-    def get_version(self, model_name: str) -> Optional[ModelVersion]:
-        """
-        Get model's latest version given the model name
-        Args:
-            model_name: model name for which the version has to be identified.
-        Returns:
-            version from mlflow ModelVersion instance
-        """
-        try:
-            return self.client.get_latest_versions(name=model_name, stages=[])[-1].version
-        except RestException as ex:
-            _LOGGER.error("Error when getting model version: %r", ex)
-            return None
+    def __delete_stale_models(self, skeys: Sequence[str], dkeys: Sequence[str]):
+        model_name = self.construct_key(skeys, dkeys)
+        list_model_versions = list(self.client.search_model_versions(f"name='{model_name}'"))
+        if len(list_model_versions) > self.models_to_retain:
+            models_to_delete = list_model_versions[self.models_to_retain :]
+            for stale_model in models_to_delete:
+                self.delete(skeys=skeys, dkeys=dkeys, version=stale_model.version)
+                _LOGGER.debug("Deleted stale model version : %s", stale_model.version)
+
+    def __load_artifacts(
+        self, skeys: Sequence[str], dkeys: Sequence[str], version_info: ModelVersion
+    ) -> Tuple[Artifact, Dict[str, Any]]:
+        model_key = self.construct_key(skeys, dkeys)
+        model = self.handler.load_model(model_uri=f"models:/{model_key}/{version_info.version}")
+        _LOGGER.info("Successfully loaded model %s from Mlflow", model_key)
+
+        run_info = mlflow.get_run(version_info.run_id)
+        metadata = run_info.data.params or {}
+        _LOGGER.info(
+            "Successfully loaded model = %s with version %s Mlflow!",
+            model_key,
+            version_info.version,
+        )
+        return model, metadata
